-- Project: GridConfigurableLayouts r69
-- File: GridConfigurableLayouts.lua
-- Last Modified: 2011-03-09T07:56:40Z
-- Author: msaint
-- Project desc: Allows in-game creation of highly customized layouts in Grid
-- File desc: All core functionality and GUI options 

local FILE_VERSION = tonumber("69") or 5000
local OUR_VERSION = tonumber("r69") or 5000

--{{{ Create the module
local GridConfigurableLayouts = Grid:GetModule("GridConfigurableLayouts", true) or Grid:NewModule("GridConfigurableLayouts", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0")
GridConfigurableLayouts.Versions = GridConfigurableLayouts.Versions or {} 
if (GridConfigurableLayouts.Versions.Base and GridConfigurableLayouts.Versions.Base >= FILE_VERSION) then return end
GridConfigurableLayouts.Versions.Base = FILE_VERSION
GridConfigurableLayouts.menuName = "GCLayouts"
GridConfigurableLayouts.Debug = Grid.Debug
---}}}


--{{{ Libraries
local Grid = Grid
local GridRoster = Grid:GetModule("GridRoster")
local GridLayout = Grid:GetModule("GridLayout")
local media = LibStub("LibSharedMedia-3.0")
--}}}


--{{{ Hack the localization function for easy implementation later
GridConfigurableLayouts.L = GridConfigurableLayouts.L or {} 
local L = GridConfigurableLayouts.L 
setmetatable(L,{ __index = function(l, text) return text end})
---}}}


---{{{ Make everthing local that we can
   --Lua library:
local ipairs, pairs, type, tinsert, tremove, table, string, math, tonumber, tostring, gmatch = ipairs, pairs, type, tinsert, tremove, table, string, math, tonumber, tostring, gmatch
local tContains = tContains -- not standard lua
   --WoW api:
local GetRealZoneText, GetRaidRosterInfo, GetNumPartyMembers = GetRealZoneText, GetRaidRosterInfo, GetNumPartyMembers 
local GetUnitName, UnitName, UnitIsInMyGuild, UnitInRaid, UnitInParty, UnitIsPlayer, UnitIsUnit, UnitName, UnitClass = GetUnitName, UnitName, UnitIsInMyGuild, UnitInRaid, UnitInParty, UnitIsPlayer, UnitIsUnit, UnitName, UnitClass
local InCombatLockdown = InCombatLockdown
local CreateFrame = CreateFrame
---}}}     


--{{{ make debug function local for convenience
local function debug(...)
   GridConfigurableLayouts:Debug("[GCL]", ...)
end 
--}}}


---{{{ File scope
-- Variables
local _
local combatLockdown = false
local updateGroupsQueued = false
local redoScheduled = nil
local updateScheduled = nil
local isRegistered = false
local defaultLayout = nil
local currentLayout = nil
-- Tables
local layoutNames = {}
local rosterZoneList = {} 
local outerFrames = {}
local outerPetFrames = {}
local emptyLayout = {
   defaults = {
  	   unitsPerColumn = 5,
	   maxColumns = 1,
	},
	[1] = {
		nameList = "",
	},
}
local filters = { --This is the order in which filters will appear in the GUI
   "tank",        --and must match the order of filterLabels defined 
   "blizztank",   --in the options section!
   "melee",
   "healer",
   "ranged",
   "guild",
   "ooz",
   "everyone",
   "rg1",
   "rg2",
   "rg3",
   "rg4",
   "rg5",
   "rg6",
   "rg7",
   "rg8",
   "pets",
   "empty",
}
---}}}


---{{{ Member variables
GridConfigurableLayouts.oozTimer = nil
GridConfigurableLayouts.newLayoutName = ""


---{{{ Utility functions
local function hasVisibleUnitFrame(f)
   -- Returns true if any child (recursively) contains a visible unit frame
   -- This function borrowed from GridLayout.lua with a small modification
	if f:IsVisible() and f:GetAttribute("unit") then
		return true
	end
   for _, f in ipairs({f:GetChildren()}) do
		if hasVisibleUnitFrame(f) then
			return true
		end
	end
end


-- Returns the maximum units allowed for partyState
local maxPartyUnits = {
   solo = 5,
	party = 5,
	raid_10 = 10,
   raid_25 = 25,
	bg = 40,
	arena = 5,
}
local function getMaxUnits(partyState)
   return maxPartyUnits[partyState]
end



local function tFind(t, val)
   -- Returns a member of t, if any, whose value is val.  Otherwise nil.
   if (t and val) then
      for member, value in pairs(t) do
         if (val == value) then
            return member
         end
      end
   end
end


local function tInsertT(t1, pos, t2)
   -- Inserts the entire contents of array t2 into array t1 at position pos
   if (type(pos) == "table") then
      t2 = pos
      pos = nil
   end 
   if (t1 and t2) then
      local val
      pos = pos or #t1 + 1
      for i = #t2, 1, -1 do
         val = t2[i]
         if (not tContains(t1, val)) then
            tinsert(t1, pos, val)
         end
      end
   end
end


local function groupSort(t, primaryKey)
   -- For 'index' or 'name', just sorts on those.  For 'class' and 'role', first
   -- sorts on 'name', then sorts on 'class' or 'role'
   if (t and primaryKey) then
      local secondaryKey = primaryKey ~= "name" and primaryKey ~= "index" and "name" or nil
      local key = secondaryKey or primaryKey
      local comp = function (t1, t2)
         return (t1[key] and t2[key] and t1[key] < t2[key])
      end
      table.sort(t, comp)
      if secondaryKey then
         key = primaryKey
         local order = currentLayout.sortOptions.sortOrder[key]
         local comp = function (t1, t2)
            return (t1[key] and t2[key] and order[t1[key]] < order[t2[key]])
         end
         table.sort(t, comp)
      end
   end
end


local function UnitIndex(unit)
   -- Gets the unit index regardless of party type
   local index = UnitInRaid(unit) 
   if (not index) then
      if UnitIsPlayer() then 
         index = 0
      else 
         for i=0, GetNumPartyMembers() do  
            if UnitIsUnit(unit, "party"..tostring(i+1)) then
               index = i + 1
               break
            end
         end
      end
   end
   return index
end
---}}}

--{{{ Update outdated db
local function fixOldDB()
   -- Placeholder for future issues
end
--}}}


--{{{ AceDB defaults
GridConfigurableLayouts.defaultDB = {
   debug = false,
   checkOOZFrequency = 3,
   useBlizzRoles = true,
   blizzRolesRule = false,
   configLayout = "GCL Default",
   filterToColor = "tank",
   filterColors = {
      tank = {r = 0.9, g = 0.5, b = 0.5, a = 0.8},
      blizztank = {r = 0.9, g = 0.5, b = 0.5, a = 0.8},
      melee = {r = 0.9, g = 0.9, b = 0.5, a = 0.8},
      healer = {r = 0.5, g = 0.9, b = 0.5, a = 0.8},
      ranged = {r = 0.5, g = 0.5, b = 0.9, a = 0.8},
      ooz = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      rg1 = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      rg2 = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      rg3 = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      rg4 = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      rg5 = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      rg6 = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      rg7 = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      rg8 = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      pets = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      empty = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
      guild = {r = 0.3, g = 0.9, b = 0.6, a = 0.5},
      everyone = {r = 0.5, g = 0.5, b = 0.5, a = 0.5},
   },
   layouts = {
      ["GCL Default"] = {
         unitsPerColumn = 5,
         createBorders = true,
         allowRepeats = false,
         numGroups = 6,
         groupType = {"tank", "melee", "healer", "ranged", "pets", "ooz", "none", "none", "none", "none", "none" },
         nameList = {"", "", "", "", "", "", "", "", "", "", "" },
         petOptions = {
            ["allPets"] = false,
            ["HUNTER"] = true,
            ["WARLOCK"] = true,
            ["DEATHKNIGHT"] = true,
         },
         sortOptions = {
            sortMethod = "name",
            sortOrder = {
               role = {"tank", "healer", "melee", "ranged"},
               class = {"deathknight", "druid", "hunter", "mage", "paladin", "priest", "rogue", "shaman", "warlock", "warrior"},
            },
         },
      },
      ["GCL Group 10"] = {
         unitsPerColumn = 5,
         createBorders = false,
         allowRepeats = false,
         numGroups = 5,
         groupType = {"tank", "rg1", "rg2", "pets", "ooz", "none", "none", "none", "none", "none", "none" },
         nameList = {"", "", "", "", "", "", "", "", "", "", "" },
         petOptions = {
            ["allPets"] = true,
            ["HUNTER"] = true,
            ["WARLOCK"] = true,
            ["DEATHKNIGHT"] = true,
         },
         sortOptions = {
            sortMethod = "role",
            sortOrder = {
               role = {"tank", "healer", "melee", "ranged"},
               class = {"deathknight", "druid", "hunter", "mage", "paladin", "priest", "rogue", "shaman", "warlock", "warrior"},
            },
         },
      },
      ["GCL Group 25"] = {
         unitsPerColumn = 5,
         createBorders = false,
         allowRepeats = false,
         numGroups = 8,
         groupType = {"tank", "rg1", "rg2", "rg3", "rg4", "rg5", "pets", "ooz", "none", "none", "none" },
         nameList = {"", "", "", "", "", "", "", "", "", "", "" },
         petOptions = {
            ["allPets"] = true,
            ["HUNTER"] = true,
            ["WARLOCK"] = true,
            ["DEATHKNIGHT"] = true,
         },
         sortOptions = {
            sortMethod = "name",
            sortOrder = {
               role = {"tank", "healer", "melee", "ranged"},
               class = {"deathknight", "druid", "hunter", "mage", "paladin", "priest", "rogue", "shaman", "warlock", "warrior"},
            },
         },
      }, 
   },
}
defaultLayout = GridConfigurableLayouts.defaultDB.layouts["GCL Default"]
--}}}


--{{{ Options Structure
do 
   local filterLabels = {   --This is the order in which filters will appear in
      L["Tanks"],           --the GUI and must match the order of filters      
      L["Blizzard Tanks"],  --defined above in the file scope variables section!                               
      L["Melee"],
      L["Healers"],
      L["Ranged"],
      L["Guild"],
      L["Out Of Zone"],
      L["Everyone"],
      L["Raid Group 1"],
      L["Raid Group 2"],
      L["Raid Group 3"],
      L["Raid Group 4"],
      L["Raid Group 5"],
      L["Raid Group 6"],
      L["Raid Group 7"],
      L["Raid Group 8"],
      L["Pets"],
      L["Namelist (or Empty)"],
   }

   local function filterLabel(filter)
      -- Returns the label for a filter based on the internal GCL name for it. 
      -- Makes some of the handlers in the options more readable, but I wish
      -- we had inline functions.
      return filterLabels[tFind(filters, filter)]
   end

   local options = {
      ["layoutOptions"] = {
         type = "group",
         name = L["Layout Options"],
         desc = L["Options that apply to a specific layout."],
         order = 25,
         disabled = InCombatLockdown,
         handler = {
            ["hidden"] = function (self, info)
               local _, _, iT = string.find(info[#info], "(%d+)")
               return (iT and tonumber(iT) > GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups) 
            end,
         },
         args = {
            ["layoutHeader"] = {
               type = "description",
               fontSize = "medium",
               name = function () return "[" .. GridConfigurableLayouts.db.profile.configLayout .. "]\n" end, 
               order = 0,
            },
            ["layoutSelectGroup"] = {
               type = "group",
               name = "Select Layout",
               inline = true,
               order = 2,
               args = {
         	      ["configLayout"] = {
                     type = "select",
                 	   name = L["Select layout to configure:"],
                 	   desc = L["Sets which layout configuration changes apply to."],
                 	   order = 2,
                 	   width = "double",
                     disabled = InCombatLockdown,
                 	   get = function () return GridConfigurableLayouts.db.profile.configLayout end,
                 	   set = function (_, v) GridConfigurableLayouts:loadNewConfig(v) end,
                 	   values = function() return layoutNames end,
                  },
                  ["configLayoutExplanation"] = {
                     type = "description",
                     name = L["\n* All changes made in the 'General', 'Group Filters', 'Name Lists', and 'Sorting' sub-pages will apply to the layout selected above.\n"],
                     order = 3,
                     width = "full",
                  },
               },
            },
            ["createRemoveGroup"] = { 
         	   type = "group",
         	   inline = true,
               name = L["Create or Remove Layout"],
         	   order = 8,
         	   args = {
               	["newLayoutName"] = {
                     type = "input",
                     name = L["Name to create or remove:"],
                     desc = L["Name of layout to create or remove (hint: type name then press <enter>)."],
                     width = "full",
                     order = 9,
                     disabled = InCombatLockdown,
                     get = function () return GridConfigurableLayouts.newLayoutName end,
                     set = function (_, v) 
                        GridConfigurableLayouts.newLayoutName = v
                     end, 
                  },
                  ["createNewLayout"] = {
                     type = "execute",
                     name = L["Create New Layout"],
                     desc = L["Create a new layout."],
                     order = 11,
                     disabled = function () return InCombatLockdown() or (not GridConfigurableLayouts.newLayoutName or GridConfigurableLayouts.newLayoutName == "" or GridLayout.layoutSettings[GridConfigurableLayouts.newLayoutName]) end,
                     func = function () GridConfigurableLayouts:createLayout(GridConfigurableLayouts.newLayoutName) end,
                  },
                  ["removeLayout"] = {
                     type = "execute",
                     name = L["Remove Layout"],
                     desc = L["Remove a layout."],
                     order = 12,
                     disabled = function () return InCombatLockdown() or (not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.newLayoutName] or GridConfigurableLayouts.newLayoutName == "GCL Default") end,
                     func = function () GridConfigurableLayouts:removeLayout(GridConfigurableLayouts.newLayoutName) end,
                  },
               },
            },
            ["generalOptions"] = {
               type = "group",
               name = L["General"],
               desc = L["Options that apply to this layout."],
               order = 25,
               disabled = InCombatLockdown,
               args = {
                  ["layoutHeader"] = {
                     type = "description",
                     fontSize = "medium",
                     name = function () return "[" .. GridConfigurableLayouts.db.profile.configLayout .. "]\n" end, 
                     order = 0,
                  },
                  ["otherOptions"] = {
                     type = "group",
                     inline = true,
                     name = "Player Group Options",
                     order = 10,
                     args = {
                        ["unitsPerColumn"] = {
                     	   type = "range",
                      		name = L["Maximum units per column (or row)"],
                       	   desc = L["Maximum number of units to display in each column or row."],
                           width = "double",
                           order = 21,
                           disabled = InCombatLockdown,
                           max = 40,
                           min = 1,
                           step = 1,
                           get = function ()	return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].unitsPerColumn end,
                     	   set = function (_, v)
                        		GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].unitsPerColumn = v
                        	   GridConfigurableLayouts:checkUpdateCurrent()
                     	   end,
                        },
                        ["allowRepeats"] = {
                       	   type = "toggle",
                          	name = L["Allow repeats"],
                          	desc = L["Show raid members in all groups that apply, instead of only once"],
                          	order = 16,
                           width = "full",
                          	disabled = InCombatLockdown,
                          	get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].allowRepeats end,
                          	set = function (_, v)
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].allowRepeats =
                                 not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].allowRepeats
                              GridConfigurableLayouts:checkUpdateCurrent()
                           end, 
                        },
                        ["offlineIsOOZ"] = {
                       	   type = "toggle",
                          	name = L["Treat offline as out of zone"],
                          	desc = L["Put raid members that are offline in out of zone group, if present."],
                          	width = "full",
                           order = 19,
                          	disabled = InCombatLockdown,
                          	get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].offlineIsOOZ end,
                          	set = function (_, v)
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].offlineIsOOZ =
                                 not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].offlineIsOOZ
                              GridConfigurableLayouts:checkUpdateCurrent()
                           end, 
                        },
                     	["createBorders"] = {
                           type = "toggle",
                           name = L["Replace Grid Border with Group Borders"],
                           desc = L["Enable creation of colored borders around groups. Uses GridLayout border texture.\nHint: play with your padding under Layout."],
                           width = "full",
                           order = 20,
                           disabled = InCombatLockdown,
                           get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].createBorders end,
                           set = function ()
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].createBorders = 
                                 not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].createBorders
                              GridConfigurableLayouts:checkUpdateCurrent()
                           end, 
                        },
                     },
                  },                  
                  ["petOptions"] = {
                     type = "group",
                     inline = true,
                     name = L["Pet Group Options"],
                     order = 20,
                     args = {
                        ["petOptionsHeader"] = { 
                           type = "header",
                           name = L["Pet Group Options"],
                           order = 3,
                           width = "full",
               	      },
                        ["enableAllPets"] = {
                           type = "toggle",
                           name = L["Display all raid pets"],
                           desc = L["Display raid pets for all pet classes."],
                           width = "double",
                           order = 4,
                           disabled = InCombatLockdown,
                           get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"] end,
                           set = function ()
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"] =
                                 not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"]
                              GridConfigurableLayouts:checkUpdateCurrent()
                           end, 
                        },
                        ["enableHunterPets"] = {
                           type = "toggle",
                           name = L["Display hunter pets"],
                           desc = L["Display raid pets for hunters."],
                           width = "double",
                           disabled = function () return InCombatLockdown() or GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"] end, 
                           order = 5,
                           get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["HUNTER"] end,
                           set = function ()
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["HUNTER"] =
                                 not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["HUNTER"]
                              GridConfigurableLayouts:checkUpdateCurrent()
                           end, 
                        },
                        ["enableWarlockPets"] = {
                           type = "toggle",
                           name = L["Display warlock pets"],
                           desc = L["Display raid pets for warlocks."],
                           width = "double",
                           disabled = function () return InCombatLockdown() or GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"] end,
                           order = 6,
                           get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["WARLOCK"] end,
                           set = function ()
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["WARLOCK"] =
                                 not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["WARLOCK"]
                              GridConfigurableLayouts:checkUpdateCurrent()
                           end, 
                        },
                        ["enableDeathknighPets"] = {
                           type = "toggle",
                           name = L["Display deathknight pets"],
                           desc = L["Display raid pets for deathknights."],
                           width = "double",
                           disabled = function () return InCombatLockdown() or GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"] end,
                           order = 7,
                           get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["DEATHKNIGHT"] end,
                           set = function ()
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["DEATHKNIGHT"] =
                                 not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["DEATHKNIGHT"]
                              GridConfigurableLayouts:checkUpdateCurrent()
                           end, 
                        },
                     },
                  },
               },
            },
            ["editGroupFilters"] = {
               type = "group",
               name = L["Group Filters"],
               desc = L["Select filters to apply to groups in the current layout."],
               order = 27,
               disabled = InCombatLockdown,
               args = {
                  ["layoutHeader"] = {
                     type = "description",
                     fontSize = "medium",
                     name = function () return "[" .. GridConfigurableLayouts.db.profile.configLayout .. "]\n" end, 
                     order = 0,
                  },
                  ["groupType1"] = {
                     type = "select",
                     name = L["Group 1:"],
                 	   desc = L["Sets contents of the first grid group."],
                     order = 2,
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                 	   get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[1]) end,
                 	   set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[1]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                 	   values = filterLabels,
                  },
                  ["groupType2"] = {
                     type = "select",
                 	   name = L["Group 2:"],
                 	   desc = L["Sets contents of the second grid group."],
                 	   order = 3,
                 	   disabled = InCombatLockdown,
                 	   hidden = "hidden",
                 	   get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[2]) end,
                 	   set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[2]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                 	   values = filterLabels,
                  },
                  ["groupType3"] = {
                     type = "select",
                 	   name = L["Group 3:"],
                 	   desc = L["Sets contents of the third grid group."],
                 	   order = 5,
                 	   disabled = InCombatLockdown,
                 	   hidden = "hidden",
                     get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[3]) end,
                 	   set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[3]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                    	values = filterLabels,
                  },
                  ["groupType4"] = {
                     type = "select",
                 	   name = L["Group 4:"],
                 	   desc = L["Sets contents of the fourth grid group."],
                 	   order = 6,
                 	   disabled = InCombatLockdown,
                 	   hidden = "hidden",
                     get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[4]) end,
                 	   set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[4]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                    	values = filterLabels,
                  },
                  ["groupType5"] = {
                     type = "select",
                 	   name = L["Group 5:"],
                 	   desc = L["Sets contents of the fifth grid group."],
                 	   order = 7,
                 	   disabled = InCombatLockdown,
                 	   hidden = "hidden",
                 	   get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[5]) end,
                 	   set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[5]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                    	values = filterLabels,
                  },
                  ["groupType6"] = {
                     type = "select",
                 	   name = L["Group 6:"],
                 	   desc = L["Sets contents of the sixth grid group."],
                 	   order = 8,
                 	   disabled = InCombatLockdown,
                 	   hidden = "hidden",
                 	   get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[6]) end,
                 	   set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[6]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                    	values = filterLabels,
                  },
                  ["groupType7"] = {
                     type = "select",
                 	   name = L["Group 7:"],
                 	   desc = L["Sets contents of the seventh grid group."],
                 	   order = 9,
                 	   disabled = InCombatLockdown,
                 	   hidden = "hidden",
                 	   get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[7]) end,
                 	   set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[7]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                    	values = filterLabels,
                  },
                  ["groupType8"] = {
                     type = "select",
                 	   name = L["Group 8:"],
                    	desc = L["Sets contents of the eighth grid group."],
                 	   order = 10,
                 	   disabled = InCombatLockdown,
                 	   hidden = "hidden",
                    	get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[8]) end,
                    	set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[8]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                    	values = filterLabels,
                  },
                  ["groupType9"] = {
                     type = "select",
                 	   name = L["Group 9:"],
                 	   desc = L["Sets contents of the ninth grid group."],
                 	   order = 11,
                 	   disabled = InCombatLockdown,
                 	   hidden = "hidden",
                 	   get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[9]) end,
                 	   set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[9]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                 	   values = filterLabels,
                  },
                  ["groupType10"] = {
                     type = "select",
                 	   name = L["Group 10:"],
                 	   desc = L["Sets contents of the tenth grid group."],
                 	   order = 12,
                 	   disabled = InCombatLockdown,
                 	   hidden = "hidden",
                 	   get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[10]) end,
                 	   set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[10]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                    	values = filterLabels,
                  },
                  ["groupType11"] = {
                     type = "select",
                 	   name = L["Group 11:"],
                 	   desc = L["Sets contents of the eleventh grid group."],
                 	   order = 13,
                 	   disabled = InCombatLockdown,
                 	   hidden = "hidden",
                 	   get = function () return tFind(filters, GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[11]) end,
                 	   set = function (_, v)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[11]= filters[v]
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                 	   values = filterLabels,
                  },
                  ["softSpacer"] = { 
                     type = "description",
                     name = "\n",
                     order = 14,
                     width = "full",
         	      },
                  ["addGroup"] = {
                     type = "execute",
                     name = L["Add a Group"],
                     desc = L["Add a group."],
                     order = 15,
                     disabled = function () return InCombatLockdown() or (GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups > 10) end,
                     func = function ()
                        local num = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups + 1
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[num] = "empty"
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups = num
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                  },
                  ["removeGroup"] = {
                     type = "execute",
                     name = L["Remove a Group"],
                     desc = L["Remove a group."],
                     order = 16,
                     disabled = function () return InCombatLockdown() or (GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups < 2) end,
                     func = function ()
                        local num = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[num] = "none"
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups = num - 1
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                  },    
               },
            },
            ["editNameLists"] = {
               type = "group",
               name = L["Name Lists"],
               desc = L["Edit name lists to apply to groups in the current layout."],
               order = 29,
               disabled = InCombatLockdown,
               args = {
                  ["layoutHeader"] = {
                     type = "description",
                     fontSize = "medium",
                     name = function () return "[" .. GridConfigurableLayouts.db.profile.configLayout .. "]\n" end, 
                     order = 0,
                  },
                  ["nameList1"] = {
                     type = "input",
           	         name = L["Group 1:"],
           	         desc = L["Sets name list for the first grid group (comma separated list)."],
                     usage = "",
           	         order = 1,
                     width = "double",
           	         disabled = InCombatLockdown,
                     hidden = "hidden",
           	         get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[1] end,
           	         set = function (_, v)
                        v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[1] = v
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                  },
                  ["addTarget1"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 1."],
                     order = 2,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[1] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[1] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[1] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["nameList2"] = {
                     type = "input",
             	      name = L["Group 2:"],
             	      desc = L["Sets name list for the second grid group (comma separated list)."],
             	      usage = "",
                     order = 3,
                     width = "double",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
             	      get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[2] end,
             	      set = function (_, v)
                         v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                         GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[2] = v
                         GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                  },
                  ["addTarget2"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 2."],
                     order = 4,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[2] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[2] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[2] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["nameList3"] = {
                     type = "input",
                     name = L["Group 3:"],
                     desc = L["Sets name list for the third grid group (comma separated list)."],
                     usage = "",
                     order = 5,
                     width = "double",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[3] end,
                     set = function (_, v)
                        v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[3] = v
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                  },
                  ["addTarget3"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 3."],
                     order = 6,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[3] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[3] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[3] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["nameList4"] = {
                      type = "input",
                      name = L["Group 4:"],
                      desc = L["Sets name list for the fourth grid group (comma separated list)."],
                      usage = "",
                      order = 7,
                      width = "double",
                      disabled = InCombatLockdown,
                     hidden = "hidden",
                      get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[4] end,
                      set = function (_, v)
                         v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                         GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[4] = v
                         GridConfigurableLayouts:checkUpdateCurrent()
                      end,
                  },
                  ["addTarget4"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 4."],
                     order = 8,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[4] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[4] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[4] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["nameList5"] = {
                      type = "input",
                      name = L["Group 5:"],
                      desc = L["Sets name list for the fifth grid group (comma separated list)."],
                      usage = "",
                      order = 9,
                      width = "double",
                      disabled = InCombatLockdown,
                     hidden = "hidden",
                      get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[5] end,
                      set = function (_, v)
                         v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                         GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[5] = v
                         GridConfigurableLayouts:checkUpdateCurrent()
                      end,
                  },
                  ["addTarget5"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 5."],
                     order = 10,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[5] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[5] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[5] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["nameList6"] = {
                      type = "input",
                      name = L["Group 6:"],
                      desc = L["Sets name list for the sixth grid group (comma separated list)."],
                      usage = "",
                      order = 11,
                      width = "double",
                      disabled = InCombatLockdown,
                     hidden = "hidden",
                      get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[6] end,
                      set = function (_, v)
                         v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                         GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[6] = v
                         GridConfigurableLayouts:checkUpdateCurrent()
                      end,
                  },
                  ["addTarget6"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 6."],
                     order = 12,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[6] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[6] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[6] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["nameList7"] = {
                      type = "input",
                      name = L["Group 7:"],
                      desc = L["Sets name list for the seventh grid group (comma separated list)."],
                      usage = "",
                      order = 13,
                      width = "double",
                      disabled = InCombatLockdown,
                     hidden = "hidden",
                      get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[7] end,
                      set = function (_, v)
                         v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                         GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[7] = v
                         GridConfigurableLayouts:checkUpdateCurrent()
                      end,
                  },
                  ["addTarget7"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 7."],
                     order = 14,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[7] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[7] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[7] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["nameList8"] = {
                      type = "input",
                      name = L["Group 8:"],
                      desc = L["Sets name list for the eighth grid group (comma separated list)."],
                      usage = "",
                      order = 15,
                      width = "double",
                      disabled = InCombatLockdown,
                     hidden = "hidden",
                      get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[8] end,
                      set = function (_, v)
                         v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                         GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[8] = v
                         GridConfigurableLayouts:checkUpdateCurrent()
                      end,
                  },
                  ["addTarget8"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 8."],
                     order = 16,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[8] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[8] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[8] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["nameList9"] = {
                      type = "input",
                      name = L["Group 9:"],
                      desc = L["Sets name list for the ninth grid group (comma separated list)."],
                      usage = "",
                      order = 17,
                      width = "double",
                      disabled = InCombatLockdown,
                     hidden = "hidden",
                      get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[9] end,
                      set = function (_, v)
                         v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                         GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[9] = v
                         GridConfigurableLayouts:checkUpdateCurrent()
                      end,
                  },
                  ["addTarget9"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 9."],
                     order = 18,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[9] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[9] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[9] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["nameList10"] = {
                      type = "input",
                      name = L["Group 10:"],
                      desc = L["Sets name list for the tenth grid group (comma separated list)."],
                      usage = "",
                      order = 19,
                      width = "double",
                      disabled = InCombatLockdown,
                      hidden = "hidden",
                      get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[10] end,
                      set = function (_, v)
                         v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                         GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[10] = v
                         GridConfigurableLayouts:checkUpdateCurrent()
                      end,
                  },
                  ["addTarget10"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 10."],
                     order = 20,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[10] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[10] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[10] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["nameList11"] = {
                      type = "input",
                      name = L["Group 11:"],
                      desc = L["Sets name list for the eleventh grid group (comma separated list)."],
                      usage = "",
                      order = 21,
                      width = "double",
                      disabled = InCombatLockdown,
                     hidden = "hidden",
                      get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[11] end,
                      set = function (_, v)
                           v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[11] = v
                           GridConfigurableLayouts:checkUpdateCurrent()
                      end,
                  },
                  ["addTarget11"] = {
                     type = "execute",
                     name = L["+Target"],
                     desc = L["Add your target to the name list for group 11."],
                     order = 22,
                     width = "half",
                     disabled = InCombatLockdown,
                     hidden = "hidden",
                     func = function ()
                        local name = GetUnitName("target")
                        if name and (UnitInParty("target") or UnitInRaid("target")) then
                           if GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[11] ~= "" then
                              name = ", " .. name
                           end
                           GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[11] =
                              GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[11] .. name
                           GridConfigurableLayouts:checkUpdateCurrent()
                        end
                     end,
                  },
                  ["softSpacer"] = {
         		      type = "description",
                     name = "\n",
                     order = 50,
                     width = "full"
         	      },
                  ["addGroup"] = {
                     type = "execute",
                     name = L["Add a Group"],
                     desc = L["Add a group."],
                     order = 52,
                     disabled = function () return InCombatLockdown() or (GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups > 10) end,
                     func = function ()
                        local num = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups + 1
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[num] = "none"
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[num] = ""
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                  },
                  ["removeGroup"] = {
                     type = "execute",
                     name = L["Remove a Group"],
                     desc = L["Remove a group."],
                     order = 53,
                     disabled = function () return InCombatLockdown() or (GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups < 2) end,
                     func = function ()
                        local num = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[num] = "none"
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[num] = ""
                        GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups = num - 1
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end,
                  },
               },
            },
         },
      },
      ["globalOptions"] = {
         type = "group",
         name = L["Global Options"],
         desc = L["Set options that apply to all layouts created with this module."],
         order = 31,
         disabled = InCombatLockdown,
         args = {
            ["roleOptions"] = {
               type = "group",
               inline = true,
               name = L["Role Options"],
               order = 10,
               args = {
                  ["useBlizzRole"] = {
                 	   type = "toggle",
                    	name = L["Use blizzard role if inspect is unavailable"],
                    	desc = L["When inspected role is unavailable, use any available blizzard roles"],
                    	width = "full",
                     order = 17,
                    	disabled = InCombatLockdown,
                    	get = function () return GridConfigurableLayouts.db.profile.useBlizzRoles end,
                    	set = function (_, v)
                        GridConfigurableLayouts.db.profile.useBlizzRoles =
                           not GridConfigurableLayouts.db.profile.useBlizzRoles
                        GridConfigurableLayouts:wipeRoles()
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end, 
                  },
                  ["blizzRoleRules"] = {
                 	   type = "toggle",
                    	name = L["Blizzard role overrides inspected role"],
                    	desc = L["If Blizzard role is set, override inspected role."],
                    	width = "full",
                     order = 18,
                    	disabled = InCombatLockdown,
                    	get = function () return GridConfigurableLayouts.db.profile.blizzRolesRule end,
                    	set = function (_, v)
                        GridConfigurableLayouts.db.profile.blizzRolesRule =
                           not GridConfigurableLayouts.db.profile.blizzRolesRule
                        GridConfigurableLayouts:wipeRoles()
                        GridConfigurableLayouts:checkUpdateCurrent()
                     end, 
                  },
               },
            },
            ["borderColorOptions"] = {
               type = "group",
               inline = true,
               name = L["Border Colors"],
               order = 20,
               args = {
                  ["borderColor"] = {
               		type = "color",
               		name = function() return L["Border Color for "] .. filterLabel(GridConfigurableLayouts.db.profile.filterToColor) end,
               		desc = L["Adjust border color and alpha."],
               		order = 62,
               		disabled = InCombatLockdown,
               		hasAlpha = true,
               		get = function ()
                 			local filter = GridConfigurableLayouts.db.profile.filterToColor
                        local settings = GridConfigurableLayouts.db.profile
                        return settings.filterColors[filter]["r"], settings.filterColors[filter]["g"], settings.filterColors[filter]["b"], settings.filterColors[filter]["a"]
               		end,
               		set = function (_, r, g, b, a)
               			local filter = GridConfigurableLayouts.db.profile.filterToColor
                        local settings = GridConfigurableLayouts.db.profile
               			settings.filterColors[filter]["r"], settings.filterColors[filter]["g"],settings.filterColors[filter]["b"],settings.filterColors[filter]["a"] = r, g, b, a
                        GridConfigurableLayouts:UpdateSize()
                     end,
                     hasAlpha = true
               	},
                  ["borderFilterSelection"] = {
                     type = "select",
                     name = L["Filter to apply border too:"],
                     desc = L["Choose filter for which to set color."],
                     order = 63,
                     disabled = InCombatLockdown,
                     get = function () return tFind(filters, GridConfigurableLayouts.db.profile.filterToColor) end,
                     set = function (_, v)
                        GridConfigurableLayouts.db.profile.filterToColor = filters[v]
                     end,
                     values = filterLabels,
                  },
               },
            },
            ["frequencyOptions"] = {
               type = "group",
               inline = true,
               name = "Update Frequency",
               order = 30,
               args = {
                  ["checkOOZFrequency"] = {
               	   type = "range",
                		name = L["Zone check frequency"],
                 	   desc = L["Frequency with which to check for raidmember zone changes."],
                     order = 69,
                     disabled = InCombatLockdown,
                     max = 20,
                     min = 0.2,
                     step = 0.1,
                     get = function ()	return GridConfigurableLayouts.db.profile.checkOOZFrequency end,
               	   set = function (_, v)
                  		GridConfigurableLayouts.db.profile.checkOOZFrequency = v
                  	   GridConfigurableLayouts:checkUpdateCurrent()
               	   end,
                  },
               },
            },
            ["spacer"] = {
         		type = "header",
         		name = L[""],
         		order = 70,
         		width = "full"
         	},
            ["resetToDefaults"] = {
               type = "execute",
               name = L["Reset all Options"],
               desc = L["Reset all settings and layouts. This will delete any new layouts you have created."],
               order = 72,
               disabled = InCombatLockdown,
               func = function () GridConfigurableLayouts:resetAllOptions() end,
            },
         },
      },
   }
   
   ---{{{ Sorting Options
   local sortMethods = {
      ["role"] = L["Role"],
      ["class"] = L["Class"],
      ["name"] = L["Name"],
      ["index"] = L["UnitID"],
   }
   local sortTypes = {
      ["class"] = {
         ["druid"] = L["Druid"],
         ["priest"] = L["Priest"],
         ["paladin"] = L["Paladin"],
         ["shaman"] = L["Shaman"],
         ["warrior"] = L["Warrior"],
         ["deathknight"] = L["Deathknight"],
         ["rogue"] = L["Rogue"],
         ["hunter"] = L["Hunter"],
         ["mage"] = L["Mage"],
         ["warlock"] = L["Warlock"],
      },
      ["role"] = {
         ["tank"] = L["Tank"],
         ["healer"] = L["Healer"],
         ["melee"] = L["Melee"],
         ["ranged"] = L["Ranged"],
      },
   }
   local sortOrderList = {}  
   local sortOptions = {
      type = "group",
      name = L["Sorting"],
      desc = L["Select how GCL sorts names within groups."],
      order = 39,
      disabled = InCombatLockdown,
      args = {
         ["layoutHeader"] = {
            type = "description",
            fontSize = "medium",
            name = function () return "[" .. GridConfigurableLayouts.db.profile.configLayout .. "]\n" end, 
            order = 0,
         },
         ["playerFirst"] = {
            type = "toggle",
            name = L["Player First"],
            desc = L["Always show the player as the first member of the group they are in."],
            width = "normal",
            order = 4,
            disabled = InCombatLockdown,
            get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.playerFirst end,
            set = function ()
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.playerFirst =
                  not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.playerFirst
               GridConfigurableLayouts:checkUpdateCurrent()
            end, 
         },
         ["nameListFirst"] = {
            type = "toggle",
            name = function()
                  local place = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.playerFirst and L["Second"] or L["First"]
                  return L["Namelist"] .. " " .. place
               end, 
            desc = L["Don not sort the namelist, put it ahead of the rest of the group."],
            width = "normal",
            disabled = InCombatLockdown,  
            order = 6,
            get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameListFirst end,
            set = function ()
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameListFirst =
                  not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameListFirst
               GridConfigurableLayouts:checkUpdateCurrent()
            end, 
         },
         ["newLine"] = {
            type = "description",
            name = "",
            width = "full",
            order = 8,
         },
         ["sortMethod"] = {
            type = "select",
            name = L["Order Players By:"],
            desc = L["Select which method should be used to order units within each group."],
            width = "normal",
            order = 10,
            get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod end,
        	   set = function (_, v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod = v
               sortOrderList = {}
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
        	   values = sortMethods,
         },
         ["sortOrderIntro"] = {
            type = "description",
            fontSize = "medium",
            hidden = function ()
               local method = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               return (method ~= "class" and method ~= "role")
            end,
            name = function ()
               local methodT = sortMethods[GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod]
               return "\n" .. L["Current"] .. " " .. methodT .. " " .. L["Order:"] .. "\n"
            end,
            width = "full",
            order = 20,
         },
         ["sortOrder"] = {
            type = "description",
            order = 22,
            width = "full",
            hidden = function ()
               local method = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               return (method ~= "class" and method ~= "role")
            end,  
            name = function ()
               local orderT, method = "", GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               if (method == "class" or method == "role") then 
                  local order = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortOrder[method] 
                  for _, val in ipairs(order) do
                     orderT =  orderT .. sortTypes[method][val] .. ", "
                  end
               end   
               return string.gsub(orderT, ", $", "")
            end,
         },
         ["buildSortOrderIntro"] = {
            type = "description",
            fontSize = "medium",
            hidden = function ()
               local method = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               return (method ~= "class" and method ~= "role")
            end,
            name = function ()
               local methodT = sortMethods[GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod]
               return "\n" .. L["New"] .. " " .. methodT .. " " .. L["Order:"]
            end,
            width = "full",
            order = 24,
         },
         ["buildSortOrder"] = {
            type = "input",
            name = "",
            order = 26,
            width = "full", 
            disabled = true,
            hidden = function ()
               local method = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               return (method ~= "role" and method ~= "class")
            end,  
            get = function ()
               local text, method = "", GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               if (method == "class" or method == "role") then
                  for _, sType in ipairs(sortOrderList) do
                     text = text .. sortTypes[method][sType] .. ", " 
                  end
               end
               return string.gsub(text, ", $", "") 
            end,
            set = function (_, v)
               sortOrderList = {}
            end,
         },
         ["buildSortOrderButtons"] = {
            type = "multiselect",
            name = "",
            order = 28,
            width = "half",
            hidden = function ()
               local method = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               return (method ~= "role" and method ~= "class")
            end,
            get = function (_, key)
               return tContains(sortOrderList, key)
            end,
            set = function (_, key, state)
               local method = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               local i = tFind(sortOrderList, key)
               if (state and not i) then
                  tinsert(sortOrderList, key)
               elseif (not state and i) then
                  tremove(sortOrderList, i)
               end
            end,
            values = function ()
               local method = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               if (method == "role" or method == "class") then
                  return sortTypes[method]
               else
                  return {}
               end
            end, 
         },
         ["buildSortOrderCommit"] = {
            type = "execute",
            name = "Save List",
            order = 30,
            hidden = function ()
               local method = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               return (method ~= "role" and method ~= "class")
            end,
            disabled = function ()
               local method = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               return ((method == "role" and #sortOrderList ~= 4) or (method == "class" and #sortOrderList ~= 10))
            end,
            func = function ()
               local uiorder = sortOrderList or {}
               local method = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortMethod
               if ((method == "role" and #uiorder == 4) or (method == "class" and #uiorder == 10)) then                        
                  local order = {}
                  for i, v in ipairs(uiorder) do order[v] = i end
                  GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].sortOptions.sortOrder[method] = order                           
                  sortOrderList = {}
                  return true
               else
                  return L["You must include all"] .. " " .. sortMethods[method] .. " " .. L["in the sort order."]
               end
            end,
         },

      },
   }
   options["layoutOptions"]["args"]["sortOptions"] = sortOptions
   ---}}}
   
   GridConfigurableLayouts.options = {
      type = "group",
      childGroups = "tree",
      name = L["GCLayouts"],
      desc = L["Options for GridConfigurableLayouts"],
      order = -1,
      args = options,
   }
end
---}}}


---{{{ Initialization, reset, etc., . . .
function GridConfigurableLayouts:OnInitialize()
   -- Makes sure our saved variables, if any, are loaded. Sets some defaults
   -- if needed.
   --
   -- The way Grid handles modules (as opposed to GridStatus, e.g.), it doesn't
   -- always register us (this is problem with the order in which addons are
   -- loaded and the timing of initialization).
   if not self.db then
      Grid:RegisterModule("GridConfigurableLayouts", GridConfigurableLayouts)
   end
   -- small changes to the db format may need correcting
   if (not self.db.Version or self.db.Version < OUR_VERSION) then
      fixOldDB()
      self.db.Version = OUR_VERSION
   end
   -- in case the user edits the savedvariables file and deletes the configLayout ... 
   if (not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout]) then
      GridConfigurableLayouts.db.profile.configLayout = "GCL Default"
   end
   self.debugFrame = GridLayout.debugFrame
	self.debugging = self.db.profile.debug
end


function GridConfigurableLayouts:OnEnable()
   -- Usually only called when addons are first loaded.  Fills some variables
   -- with saved data from the db.
   self:loadNewConfig(self.db.profile.configLayout)
   for layoutName, _ in pairs(self.db.profile.layouts) do
      layoutNames[layoutName] = layoutName
   end
   self:addLayouts()
   self:Hook(GridLayout, "ReloadLayout")
   GridLayout:ReloadLayout()
end


function GridConfigurableLayouts:OnDisable()
   -- A little cleanup if the addon is disabled
   self:Unhook(GridLayout, "ReloadLayout")
   self:doUnregister()
   self:setCreateBorders(false)
end


function GridConfigurableLayouts:Reset()
   GridLayout:ReloadLayout()
end
---}}}


---{{{ Registrations and Callback functions
function GridConfigurableLayouts:doRegister()
   -- Called by IsaActive() when the layout is set to one that is ours
   self:RegisterMessage("Grid_EnteringCombat", GridConfigurableLayouts.setCombatLockdown)
   self:RegisterMessage("Grid_LeavingCombat", GridConfigurableLayouts.removeCombatLockdown)
   self:RegisterMessage("Grid_RosterUpdated", GridConfigurableLayouts.UpdateGroups)
   self:RegisterRoleEvents()
   if not combatLockdown and self:checkUsingOOZ() then
      self.oozTimer = self:ScheduleRepeatingTimer(GridConfigurableLayouts.OOZCheck, GridConfigurableLayouts.db.profile.checkOOZFrequency )
   end   
   isRegistered = true
end

         
function GridConfigurableLayouts:doUnregister()
   -- Called by IsActive() when the current layout changes to one that is not
   -- ours.  Also called if the addon is disabled.
   self:UnregisterMessage("Grid_EnteringCombat")
   self:UnregisterMessage("Grid_LeavingCombat")
   self:UnregisterMessage("Grid_RosterUpdated")
   self:UnregisterRoleEvents()
   self:CancelTimer(self.oozTimer, true)
   isRegistered = false
end


function GridConfigurableLayouts:setCombatLockdown()
   -- Sets the combout lockdown flag.  All updates will be queued until
   -- combat ends.
   combatLockdown = true
   GridConfigurableLayouts:CancelTimer(self.oozTimer, true)
end


function GridConfigurableLayouts:removeCombatLockdown()
   -- Puts this module back in action after combat ends, initiates updates
   -- that were postponed due to combat  
   combatLockdown = false
   if GridConfigurableLayouts:isActive() and GridConfigurableLayouts:checkUsingOOZ() then
      GridConfigurableLayouts:ScheduleRepeatingTimer("OOZCheck", GridConfigurableLayouts.db.profile.checkOOZFrequency )
   end
   if updateGroupsQueued then
      GridConfigurableLayouts:UpdateGroups()
   end
end    
   
    
function GridConfigurableLayouts:ReloadLayout()
   -- Hook on the GridLayout method (OnEnable)
   if combatLockdown then
      updateGroupsQueued = true
   else
      if GridConfigurableLayouts:isActive() then
         if (not GridConfigurableLayouts:TimeLeft(updateScheduled)) then
            if redoScheduled then
               GridConfigurableLayouts:CancelTimer(redoScheduled, true)
               redoScheduled = nil
            end
            updateScheduled = GridConfigurableLayouts:ScheduleTimer("updateLayout", 0.11)
         end
      end                                           
	end
end


function GridConfigurableLayouts:UpdateGroups()
   -- Our own function, similar to ReloadLayout, but only rebuilds groups
   -- without recreating the layout
   if combatLockdown then
      updateGroupsQueued = true
   else
      if GridConfigurableLayouts:isActive() then
         local scheduled = updateScheduled or redoScheduled 
         if (not GridConfigurableLayouts:TimeLeft(scheduled)) then
            redoScheduled = GridConfigurableLayouts:ScheduleTimer("redoGroups", 0.11) 
         end
      end                                           
	end
end


function GridConfigurableLayouts:UpdateSize()
   -- Hook on the GridLayout method, used to place group borders whenever
   -- GridLayout refreshes it's own outer border (hook set by setCreateBorders)
   if GridConfigurableLayouts:isActive() and GridConfigurableLayouts.db.profile.layouts[GridLayout.db.profile.layouts[GridRoster:GetPartyState()]].createBorders then
      GridConfigurableLayouts:makeBorders()
	end
end   


function GridConfigurableLayouts:PlaceGroup(context, layoutGroup, groupNumber)
   -- Hook on the GridLayout method, used when we need to adjust group placement
   -- to make room for group borders (hook set by setCreateBorders)
   if GridConfigurableLayouts:isActive() and GridConfigurableLayouts.db.profile.layouts[GridLayout.db.profile.layouts[GridRoster:GetPartyState()]].createBorders then
      GridConfigurableLayouts:adjustGroup(layoutGroup, groupNumber)
	end
end   


function GridConfigurableLayouts:OOZCheck()
   -- Periodically checks if any raid member has changed zones
   local pstate = GridRoster:GetPartyState()
   if not combatLockdown and pstate ~= "solo" and pstate ~= "party"  then
      if GridConfigurableLayouts:isActive() then
         for guid, unit in GridRoster:IterateRoster() do
            if (string.sub(unit, 1, 7) ~= "raidpet") then
               local zone = GetRealZoneText()
               if (string.sub(unit, 1, 4) == "raid") then
                  local raidIndex = string.sub(unit, 5)
                  local _, _, _, _, _, _, temp_zone = GetRaidRosterInfo(raidIndex)
                  zone = temp_zone or zone
      	         if rosterZoneList[guid] ~= zone and (zone ~= "Offline" or currentLayout.offlineIsOOZ) then
                     GridConfigurableLayouts:UpdateGroups()
                     return
                  end
               end
            end
         end
      end
   end
end   
---}}}


---{{{ Core functions
function GridConfigurableLayouts:createLayout(name)
   -- Occurs when "Create New Layout" button is pressed, creates a copy of
   -- GCL_Default with a new name
   if name and name ~= "" and GridLayout.layoutSettings[name] == nil then
      GridLayout:AddLayout(name, emptyLayout)
      self.db.profile.layouts[name] = defaultLayout
      layoutNames[name] = name
      self:loadNewConfig(name)
      self.newLayoutName = ""
      self:checkUpdateCurrent()
   end
end


function GridConfigurableLayouts:removeLayout(name)
   -- Occurs when "Remove Layout" button is pressed
   if self.db.profile.layouts[name] and name ~= "GCL Default" then
      GridLayout.layoutSettings[name] = nil
    	for _, party_type in ipairs(GridRoster.party_states) do
    		for _, layoutName in pairs(GridLayout.options.args[party_type .. "layout"].values) do
    		   if name == layoutName then
               GridLayout.options.args[party_type .. "layout"].values[layoutName] = nil
            end
    		end
    	end
      self.db.profile.layouts[name] = nil
      layoutNames[name] = nil
      if self.db.profile.configLayout == name then
         self:loadNewConfig("GCL Default")
      end
      self.newLayoutName = ""
      self:checkUpdateCurrent()
   end
end   
   

function GridConfigurableLayouts:setCreateBorders(doCreate)
   -- Called when creation of group borders is enabled or disabled
   if doCreate then
      if not self:IsHooked(GridLayout, "UpdateSize") then
         self:SecureHook(GridLayout, "UpdateSize")
      end
      if not self:IsHooked(GridLayout, "PlaceGroup") then
         self:SecureHook(GridLayout, "PlaceGroup")
      end
      GridLayout.frame:DisableDrawLayer("BORDER")
      GridLayout.frame:DisableDrawLayer("BACKGROUND")
   else
      if self:IsHooked(GridLayout, "PlaceGroup") then
         self:Unhook(GridLayout, "PlaceGroup")
      end
      if self:IsHooked(GridLayout, "UpdateSize") then
         self:Unhook(GridLayout, "UpdateSize")
      end
      for _, f in pairs(outerFrames) do
         f:Hide()
      end
      for _, f in pairs(outerPetFrames) do
         f:Hide()
      end
      GridLayout.frame:EnableDrawLayer("BORDER")
      GridLayout.frame:EnableDrawLayer("BACKGROUND")
   end
end


function GridConfigurableLayouts:resetAllOptions()
   --Returns the whole module to factory settings
   for layoutName, _ in pairs(self.db.profile.layouts) do
      self:removeLayout(layoutName)
   end
   self:doUnregister()
   self.db:ResetProfile()
   layoutNames = {}
   for layoutName, _ in pairs(self.db.profile.layouts) do
      layoutNames[layoutName] = layoutName
   end
   self:loadNewConfig("GCL Default")
   self:addLayouts()
   GridLayout:ReloadLayout()
end


function GridConfigurableLayouts:redoGroups()
   -- The core function of this module: processes the roster and builds the
   -- groups, which it feeds to GridLayout
   redoScheduled = nil
   local groupList = {{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {} }
   local nameLists = {}
   local oozList = {}
   local usingooz = self:checkUsingOOZ()
   local revisedNameList = {}
   local nameListed = {}
   local playerGroup
   
   -- make a clean copy of namelists with no bogus names
   -- (because names that are not in raid can block legit names from appearing)
   for i = 1, currentLayout.numGroups do
      local nlT = string.gsub(currentLayout.nameList[i], ",", " ")
      local tempnl = {}
      for name in gmatch(nlT, "(%w+)") do
         local guid = GridRoster:GetGUIDByFullName(name)
         if (guid and not nameListed[name]) then
            local unit = GridRoster:GetUnitidByGUID(guid)
            local role = self:GetRole(guid, unit)
            local index = UnitIndex(unit)
            local class =  UnitClass(unit)
            class = string.lower(class)
            nameListed[name] = true
            tinsert(tempnl, {name = name, class = class, role = role, index = index})
         end
      end
      revisedNameList[i] = tempnl
   end
   -- iterate thru the roster and place units in groups
   for guid, unit in GridRoster:IterateRoster() do
      if string.find(unit, "pet") then break end
      local name, raidgroup, isMT, zone, index = self:getUnitInfo(unit, guid)
      if (nameListed[name] and not currentLayout.allowRepeats) then break end
      local role = self:GetRole(guid, unit)
      local _, class = UnitClass(unit)
      class = string.lower(class)
      -- first check if this unit belongs in an ooz group
      if usingooz then 
         if zone ~= "Offline" then
            rosterZoneList[guid] = zone
         elseif not currentLayout.offlineIsOOZ then
            zone =  rosterZoneList[guid]
         end
         if zone ~= GetRealZoneText() and (currentLayout.allowRepeats or not nameListed[name]) then
            tinsert(oozList, {name = name, class = class, role = role})
            nameListed[name] = true
         end
      end
      -- then fill the groups in order
      for i = 1, currentLayout.numGroups do
         if (currentLayout.groupType[i] == "everyone") or 
               (role == currentLayout.groupType[i]) or
               (currentLayout.groupType[i] == "blizztank" and isMT) or
               (currentLayout.groupType[i] == raidgroup) or
               (currentLayout.groupType[i] == "guild" and UnitIsInMyGuild(unit))
               then
            if (currentLayout.sortOptions.playerFirst and UnitIsPlayer(unit)) then
               playerGroup = i -- it will be inserted later
            else
               tinsert(groupList[i], {name = name, class = class, role = role, index = index})
            end
            nameListed[name] = true
         end
      end
   end
   -- if there is an ooz group, fill it
   if usingooz then
      for i = 1, currentLayout.numGroups, 1 do
         if currentLayout.groupType[i] == "ooz" then
            tInsertT(groupList[i], oozList)
         end
      end
   end
   -- add namelists to the groups (if they are supposed to be sorted)
   if (not currentLayout.sortOptions.nameListFirst) then
      for i = 1, currentLayout.numGroups do
         if revisedNameList[i] ~= {} then
            tInsertT(groupList[i], revisedNameList[i])
         end
      end
   end
   -- sort the groups
   for i = 1, currentLayout.numGroups do
      groupSort(groupList[i], currentLayout.sortOptions.sortMethod)
   end
   -- Add the namelists now if they are not to be sorted
   if currentLayout.sortOptions.nameListFirst then
      for i = 1, currentLayout.numGroups do
         if revisedNameList[i] ~= {} then
            tInsertT(groupList[i], 1, revisedNameList[i])
         end
      end
   end
   -- If the player is to be first in her group, add her now
   if currentLayout.sortOptions.playerFirst and playerGroup then
      tinsert(groupList[playerGroup], 1, {name = UnitName("player")})
   end
   -- create the namelist strings
   for i = 1, currentLayout.numGroups do
      nameLists[i] = ""
      for _, entry in ipairs(groupList[i]) do
         nameLists[i] = nameLists[i] .. "," .. entry.name
      end
   end
   -- update the nameLists in the actual GridLayout headers
   do
      local i = 1
      for j = 1, currentLayout.numGroups do
         if currentLayout.groupType[j] ~= "pets" then
            local header = GridLayout.layoutGroups[i]
            if header then
               header:SetAttribute("nameList", nameLists[j])
               i = i + 1
            end   
         end
      end   
   end
end
   
function GridConfigurableLayouts:updateLayout()
   -- Rebuilds the current layout and inserts it into GridLayout
   updateScheduled = nil
   local partyState = GridRoster:GetPartyState() 
   local layoutName = GridLayout.db.profile.layouts[partyState]
   local unitsPerColumn = currentLayout.unitsPerColumn
   local maxColumns = math.ceil(getMaxUnits(partyState) / unitsPerColumn) 
   local petGroupFilter = ""
   
   -- create groupFilter for pet group
   if not currentLayout.petOptions["allPets"] then
      for class, enabled in pairs(currentLayout.petOptions) do
         if class ~= "allPets" and enabled == true then
            petGroupFilter = petGroupFilter .. "," .. class
         end
      end
   end
   
   -- pass the updated layout to GridLayout
   local realIndex = 1
   for i = 1, 11, 1 do
      GridLayout.layoutSettings[layoutName][i] = nil
      if currentLayout.groupType[i] ~= "none" then
         GridLayout.layoutSettings[layoutName][realIndex] = {}
         GridLayout.layoutSettings[layoutName][realIndex]["unitsPerColumn"] = unitsPerColumn
         GridLayout.layoutSettings[layoutName][realIndex]["maxColumns"] = maxColumns
         if currentLayout.groupType[i] == "pets" then
            GridLayout.layoutSettings[layoutName][realIndex]["isPetGroup"] = true
            GridLayout.layoutSettings[layoutName][realIndex]["groupFilter"] = "1,2,3,4,5,6,7,8"
            GridLayout.layoutSettings[layoutName][realIndex]["nameList"] = ""
            if not currentLayout.petOptions["allPets"] then
               GridLayout.layoutSettings[layoutName][realIndex]["groupFilter"] = petGroupFilter
            end
         else
            GridLayout.layoutSettings[layoutName][realIndex]["isPetGroup"] = false
            GridLayout.layoutSettings[layoutName][realIndex]["groupFilter"] = nil
            GridLayout.layoutSettings[layoutName][realIndex]["sortMethod"] = "NAMELIST"
         end
      realIndex = realIndex + 1
      end
   end
   -- update namelists
   self:redoGroups()
end


function GridConfigurableLayouts:makeBorders()
   -- Implements group borders around existing groups
   local groupCount = 1
   local groupColors, petGroupColors, totalGroups = self:getGroupColors()
   local borderTexture = media:Fetch(media.MediaType.BORDER, GridLayout.db.profile.borderTexture)
   
   for _, f in pairs(outerFrames) do
      f:Hide()
   end
   
   for _, f in pairs(outerPetFrames) do
      f:Hide()
   end
   
   for i, layoutGroup in ipairs(GridLayout.layoutGroups) do
		if layoutGroup:IsVisible() and hasVisibleUnitFrame(layoutGroup) then
         local f = outerFrames[i] or CreateFrame("Frame", "GCLOuterFrame" .. i, GridLayout.frame)
         local x, y = layoutGroup:GetLeft() - GridLayout.frame:GetLeft(), layoutGroup:GetTop() - GridLayout.frame:GetTop()
         local width, height = layoutGroup:GetWidth(), layoutGroup:GetHeight()
         x = x - 5
         y = y + 5
         width = width + 10
         height = height + 10
         f:SetWidth( width)
         f:SetHeight( height)
         f:SetPoint("TOPLEFT", x, y)
         f:SetBackdrop({
				bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
				edgeFile = borderTexture, edgeSize = 16,
				insets = {left = 4, right = 4, top = 4, bottom = 4},
			})
         local color = groupColors[i]
         f:SetBackdropBorderColor(color.r, color.g, color.b, color.a)
			f:SetBackdropColor(0, 0, 0, 0)
			f:Show()
			outerFrames[i] = f
		end
	end
   for i, layoutGroup in ipairs(GridLayout.layoutPetGroups) do
		if layoutGroup:IsVisible() and hasVisibleUnitFrame(layoutGroup) then
         local f = outerPetFrames[i] or CreateFrame("Frame", "GCLOuterPetFrame" .. i, GridLayout.frame)
         local x, y = layoutGroup:GetLeft() - GridLayout.frame:GetLeft(), layoutGroup:GetTop() - GridLayout.frame:GetTop()
         local width, height = layoutGroup:GetWidth(), layoutGroup:GetHeight()
         x = x - 5
         y = y + 5
         width = width + 10
         height = height + 10
         f:SetWidth( width)
         f:SetHeight( height)
         f:SetPoint("TOPLEFT", x, y)
         f:SetBackdrop({
				bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
				edgeFile = borderTexture, edgeSize = 16,
				insets = {left = 4, right = 4, top = 4, bottom = 4},
			})
         local color = petGroupColors[i]
         f:SetBackdropBorderColor(color.r, color.g, color.b, color.a)
			f:SetBackdropColor(0, 0, 0, 0)
			f:Show()
			outerPetFrames[i] = f
		end
	end
end


function GridConfigurableLayouts:adjustGroup(layoutGroup, groupNumber)
   -- For when group borders are active: Adjusts the placement of the group frames
   -- created by GridLayout to make room for our group borders
   local f = layoutGroup
   local GLSettings = GridLayout.db.profile
   local horizontal = GLSettings.horizontal
   local groupAnchor = GLSettings.groupAnchor
   
   local adjustX, adjustY = 0, 0
   
   if horizontal then
      adjustX = 0
      if groupAnchor == "TOPLEFT" or groupAnchor == "TOPRIGHT" then
         adjustY = -10
      else
         adjustY = 10 
      end
   else
      adjustY = 0
      if groupAnchor == "TOPLEFT" or groupAnchor == "BOTTOMLEFT" then
         adjustX = 10
      else
         adjustX = -10
      end
   end
          
   local p, rT, rP, x, y = f:GetPoint(1)
   f:SetPoint(p, rT, rP, x + adjustX, y + adjustY) 
end
---}}}


---{{{ other functions
function GridConfigurableLayouts:isActive()
   -- Checks if the active layout is one of ours, and if so, makes sure we are
   -- registered to react to everything we need to track
   local party_type = GridRoster:GetPartyState()
   currentLayout = self.db.profile.layouts[GridLayout.db.profile.layouts[party_type]]
   self:setCreateBorders(currentLayout and currentLayout.createBorders)
   if currentLayout and not isRegistered then
      self:doRegister()
   elseif not currentLayout and isRegistered then
      self:doUnregister()
   end
   return (not (not currentLayout))
end


function GridConfigurableLayouts:checkUpdateCurrent()
   -- When a layout is modified in options, check if the config layout is the
   -- currently active layout and, if so, reload it with the modified settings
   local party_type = GridRoster:GetPartyState()
   if GridLayout.db.profile.layouts[party_type] == self.db.profile.configLayout then
      self:CancelTimer(self.oozTimer, true)
      if not combatLockdown and self:checkUsingOOZ() then
         self.oozTimer = self:ScheduleRepeatingTimer(GridConfigurableLayouts.OOZCheck, GridConfigurableLayouts.db.profile.checkOOZFrequency )                                                                                                                                                                       
      end
      GridLayout:ReloadLayout()
   end
end            


function GridConfigurableLayouts:getUnitInfo(unit, guid)
   -- Returns unit info used by redoGroups
   local raidIndex, name = UnitIndex(unit), GridRoster:GetFullNameByGUID(guid) 
   local raidgroup, isMT = "rg1", false
   local zone = GetRealZoneText() -- note: we can't get zone info for party members
   if UnitInRaid(unit) then
      local temp_zone, blizzRole
      _, _, raidgroup, _, _, _, temp_zone, _, _, blizzRole = GetRaidRosterInfo(raidIndex)
      zone = temp_zone or zone -- this deals with a Blizz bug where reloading ui causes GetRaidRosterInfo to return empty zone info for player
      isMT = (blizzRole == "MAINTANK")
      raidgroup = "rg" .. raidgroup
   end
   return name, raidgroup, isMT, zone, raidIndex
end


function GridConfigurableLayouts:checkUsingOOZ()
   -- Checks if there is an "Out of zone" group in the current layout
   for _, groupType in ipairs(currentLayout.groupType) do
      if groupType == "ooz" then
         return true
      end
   end
   return false
end


function GridConfigurableLayouts:getGroupColors()
   -- Used by MakeBorders, this creates tables of group colors that match
   -- the group borders and pet group borders in order of their creation 
   local settings = self.db.profile
   local i, j, k = 0, 0, 0
   local groupColors = {}
   local petGroupColors = {}
   for _, groupType in pairs(currentLayout.groupType) do
      if groupType == "pets" then
         j = j + 1
         k = k + 1
         local gc = settings.filterColors["pets"] 
         petGroupColors[j] =  {r = gc.r, g = gc.g, b = gc.b, a = gc.a}
         petGroupColors[j]["realIndex"] = k
      elseif groupType == "none" then
         break
      else
         i = i + 1
         k = k + 1
         local gc = settings.filterColors[groupType] 
         groupColors[i] =  {r = gc.r, g = gc.g, b = gc.b, a = gc.a}
         groupColors[i]["realIndex"] = k
      end
   end
   return groupColors, petGroupColors, i + j
end


function GridConfigurableLayouts:addLayouts()
   -- Adds placeholders for all our layouts, so that they are visible to GridLayout
   for _, name in pairs(layoutNames) do
      if GridLayout.layoutSettings[name] == nil then
         GridLayout:AddLayout(name, emptyLayout)
      end
   end
end


function GridConfigurableLayouts:loadNewConfig(layoutName)
   -- When we choose a new layout to configure, we have a few things to do
   self.db.profile.configLayout = layoutName
   if not self.db.profile.layouts[layoutName].numGroups then
      for i = 11, 1, -1 do
         if self.db.profile.layouts[layoutName].groupType[i] ~= "none" then
            self.db.profile.layouts[layoutName].numGroups = i
            break
         end
      end
   end
end
---}}}